/*!
 * @file MoveWheelChair.cpp
 * @brief 全方位移動搭乗型ロボットプラットフォームOmni-Wheelを制御するためのクラス
 *
 * @author Masaharu Shimizu@Future Robotics Technology Center, Chiba Institute of Technology
 *
 * @date 2008/07/16  林原先生作成のソースをlinuxへ移植
 * @date 2008/07/29  オドメトリ計算で使用される時間を定数から計測時間に変更
 * @date 2009/06/26  オドメトリ計算において回転角度が考慮されていなかったため、X,Yにsin,cosを乗ずる
 *                   また、回転方向が右回りが正だったので左回りを正とする
 *                   CalcOdometryのodometry引数を削除
 * @date 2009/08/20 x,yで速度指令にたいしての速度が異なっており、y方向の進みが遅いため1.25倍して補足する
 * @date 2009/08/20 x,yのオドメトリで車輪回転数をベースに距離を求めているためy方向が短くなる。そのため1.25倍して補足する
 * @date 2010/06/29 CANインターフェースをInterface社IFCAN以外に、PeakSystem社PCAN,CANopenミドルウェアCanfestivalに対応
 * @date 2010/07/02 GetEncoder,Recvの返値を変更
 *
 */

#include "MoveWheelChair.h"
#include "timespec.h"
#include <time.h>
#include <string.h>
#include <stdio.h>
#include <iostream>
#include <iomanip>

#define GEAR_RATIO		42.0f
#define WHEEL_DIAMETER	0.25f    //!< 車輪直径
#define WHEEL_LENGTH	0.6794f  //!< 対角車輪距離

//#define DEBUG

CMoveWheelChair::CMoveWheelChair()
{
	gain_x = 1.0;
	gain_y = 1.0;
	gain_theta = 1.0;
}

CMoveWheelChair::~CMoveWheelChair()
{
}

/**
 * CANインターフェースの初期化
 */
int CMoveWheelChair::Init(void)
{
	int i;
	int m_can_port = 32;

	for(i = 0; i < 3; i ++) odometry[i] = 0.0f;


#ifdef IFCAN
	char portString[10];
	strcpy(portString, "ifcan1");

    if((can_fd = CanOpenPort(portString)) < 0){
    	printf("Failed to open %s\n", portString);
    	return -1;
    }
    CAN_PORT_CONFIG Conf;

    // モードとボーレートの設定
    CanGetConfig(can_fd, &Conf);
    //Conf.ulLineMode = CAN_NORMAL_MODE;
    Conf.ulBaudRate = CAN_BAUDRATE_1M;
    CanSetConfig(can_fd, &Conf);

    usleep(1000);

	// CANをアクティブにする
	CanActivate(can_fd);
#endif

#ifdef CANFESTIVAL
    s_BOARD board = {"32", "1M"};
    if((can_fd = canOpen_driver(&board))==0){
    	std::cout << "Can not open can_port[" << m_can_port << "]" << std::endl;
    	return 1;
    }
#endif

#ifdef PCAN
    char busname[64];
    sprintf(busname,"/dev/pcan%s","32");
    can_fd = LINUX_CAN_Open(busname, O_RDWR);
    if(can_fd && CAN_BAUD_1M) {
    	CAN_Init(can_fd, CAN_BAUD_1M, CAN_INIT_TYPE_ST);
    	//std::cout << "CAN INIT";
    }
    else{
    	fprintf(stderr, "PCAN (Peak_Linux) : error opening %s\n", busname);
    }
#endif

	usleep(1000);

	SetOdometryGain(1.0, 1.0, 1.0);
	// FTMD電源投入時にエラー状態になることがあるのでサーボOFF、ON、OFFを実行
	Servo(0);
	usleep(1000);
	Servo(1);
	usleep(1000);
	Servo(0);

	return 0;
}

/**
 * 車体の停止命令とCANインターフェースの終了処理
 */
int CMoveWheelChair::Close()
{
	// 停止
	char com[] = {0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00};
	Send(0x201, sizeof(com), com);
	usleep(1000);

	// クローズ
#ifdef IFCAN
	CanDeactivate(can_fd);
	usleep(1000);
	CanClosePort(can_fd);
#endif

#ifdef CANFESTIVAL
	canClose_driver(can_fd);
#endif

#ifdef PCAN
	CAN_Close(can_fd);
#endif
	return 0;
}

#define MAX_VELOCITY	2.0f
#define MAX_ANG_VELO	2.0f

/**
 * モーターに速度指令を送信する
 * @param[in] x x軸方向の速度 (m/s)
 * @param[in] y y軸方向の速度 (m/s)
 * @param[in] z ヨー軸方向の角速度 (rad/s)
 * @return 成功の場合は0
 */
int CMoveWheelChair::Move(float x, float y, float z)
{
	char com[8];
	int q[4], i;
	int enc[4] = {0,0,0,0};
	static float xd = 0.0f, yd = 0.0f, zd = 0.0f;
	static struct timespec now, old, period;
	static int count_first = 0;
	float xdt, ydt, zdt;
	float a = 0.60f;


	//最大値、最小値に制限
	if (x < -MAX_VELOCITY) x = -MAX_VELOCITY;
	else if (x > MAX_VELOCITY) x = MAX_VELOCITY;
	if (y < -MAX_VELOCITY) y = -MAX_VELOCITY;
	else if (y > MAX_VELOCITY) y = MAX_VELOCITY;
	if (z < -MAX_ANG_VELO) z = -MAX_ANG_VELO;
	else if (z > MAX_ANG_VELO) z = MAX_ANG_VELO;

	//前回の実行値と今回の指令値の重み付き平均
	xd = a * xd + (1.0f - a) * x;
	yd = a * yd + (1.0f - a) * y;
	zd = a * zd + (1.0f - a) * z;

	//速度をモータ回転数に変換
	xdt = xd / (WHEEL_DIAMETER * M_PI) * GEAR_RATIO * 60.0f;
	ydt = yd / (WHEEL_DIAMETER * M_PI) * GEAR_RATIO * 60.0f * 1.25; //y方向の進みが遅いため1.25倍して補足する

	zdt = zd * (WHEEL_LENGTH / 2.0f) / 0.732f / (WHEEL_DIAMETER * M_PI) * GEAR_RATIO * 60.0f;


	q[0] = -xdt + ydt - zdt;
	q[1] =  xdt + ydt - zdt;
	q[2] = -xdt - ydt - zdt;
	q[3] =  xdt - ydt - zdt;

	for(i = 0;i < 4;i ++){
		com[i*2  ] =  q[i]       & 0xFF;
		com[i*2+1] = (q[i] >> 8) & 0xFF;
	}

	if (!GetEncoder(enc)){
		clock_gettime(CLOCK_REALTIME, &now);
		if(count_first == 0){
			old = now;
			count_first = 1;
		}
		period  = now;
		timespec_sub(period, old);
		old = now;

		//std::cout << "period:" << timespec2micro(period) << "[us]";
		//std::cout << count << " " << timespec2micro(period) << std::endl;
		CalcOdometry(enc, (float)(timespec2micro(period)/1000000.0f));

	}

	Send(0x201, sizeof(com), com);
	return 0;
}
/**
 * @braf　オドメトリゲインを設定する
 */
int CMoveWheelChair::SetOdometryGain(double g_x, double g_y, double g_theta){
	gain_x = g_x;
	gain_y = g_y;
	gain_theta = g_theta;

	return 0;
}

/**
 * オドメトリを計算する
 * @param[in]  enc[4] 各車輪の回転速度[rad/s]
 * @param[in]  st 速度を位置に積算するためのサンプリング時間
 * @return 成功の場合は0
 */
int CMoveWheelChair::CalcOdometry(int enc[4], float st){
	float x=0.0, y=0.0; //!<ロボット中心に見た場合の1サンプリングあたりの移動距離
	x = (- enc[0] + enc[1] - enc[2] + enc[3]) / 4.0f * (WHEEL_DIAMETER * M_PI) / GEAR_RATIO / 60.0f * st * gain_x;
	//gain_y = 0.8;
	y = (+ enc[0] + enc[1] - enc[2] - enc[3]) / 4.0f * (WHEEL_DIAMETER * M_PI) / GEAR_RATIO / 60.0f * st * gain_y * 0.8;

	//thetaの回転方向を左回りを正にするためにマイナスする
	//odometry[2] -= (float)(+ enc[0] + enc[1] + enc[2] + enc[3]) / 4.0f / (WHEEL_LENGTH / 2.0f) * 0.732f * WHEEL_DIAMETER * M_PI / GEAR_RATIO / 60.0f * st;
	//gain_theta = 0.65;
	odometry[2] -= (float)(+ enc[0] + enc[1] + enc[2] + enc[3]) / 4.0f / (WHEEL_LENGTH / 2.0f) * WHEEL_DIAMETER * M_PI / GEAR_RATIO / 60.0f * st * gain_theta * 0.65;

	//回転角度を考慮してオドメトリを計算する
	odometry[0] += cos(odometry[2]) * x - sin(odometry[2]) * y;
	odometry[1] += sin(odometry[2]) * x + cos(odometry[2]) * y;

#ifdef DEBUG
	//std::cout.width(6);
	//std::cout.setf(std::ios::right, std::ios::adjustfield);
	std::cout << " st:" << std::setw(6) << st*1000 << "[ms] ";
	for(int i=0; i<4; i++)
		std::cout << "enc[" << i << "]:" << std::setw(4) << enc[i] << "[rpm] ";
	std::cout << std::endl;
	std::cout << " odo[x]=" << std::setw(6) << odometry[0] << "[m]";
	std::cout << " odo[y]=" << std::setw(6) << odometry[1] << "[m]";
	std::cout << " odo[z]=" << std::setw(6) << odometry[2] << "[rad] " << odometry[2]/M_PI * 180 << "[degree]";
	std::cout << std::endl;
	//std::cout << "\033[2J";
	//std::cout << "\r";
#endif
	return 0;
}

/**
 * モーターからオドメトリ情報を取得する
 * @param[out] odo オドメトリ座標系での座標x(m), y(m), theta(rad)
 * @return 成功の場合は0
 */
int CMoveWheelChair::GetOdometry(float odo[3]){
	int i;
	for(i = 0;i < 3;i ++) odo[i] = odometry[i];

	return 0;
}

/**
 * オドメトリ情報を設定する
 * @param[in] odo オドメトリ座標系での座標x(m), y(m), theta(rad)
 * @return 成功の場合は0
 */
int CMoveWheelChair::SetOdometry(float odo[3]){
	int i;
	for(i = 0;i < 3;i ++) odometry[i] = odo[i];

	return 0;
}

/**
 * 車輪の回転速度情報を取得する
 * @param[out] enc　各車輪の合計４輪の回転速度
 * @return 成功の場合は0
 */
int CMoveWheelChair::GetEncoder(int enc[4])
{
	int id, no, measure[4];
	unsigned long msgCount;

	CanMessage  *recvMsg;
	recvMsg = new CanMessage[RECV_BUFF_SIZE];

	if (!Recv(&msgCount, recvMsg)){
		//std::cout << "msgCount=" << msgCount ;

		for(unsigned int i = 0; i < msgCount; i++){
			id = recvMsg[i].cob_id;
			if ((id > 0x380)&&(id <= 0x384)){
				no = id - 0x381;
				measure[no] = (short)(recvMsg[i].data[7]<<8)+recvMsg[i].data[6]; // rpm
				enc[no] = measure[no];
/*
 				printf("id=0x%x msg.length=%d", id, recvMsg[i].len);
				std::cout << " enc[" << no << "]=" << enc[no] << "[rpm] ";
				printf(" bData[7]:%x bData[6]:%x\n",recvMsg[i].data[7], recvMsg[i].data[6]);
*/
			}
		}
		delete [] recvMsg;
		return 0;
	}
	delete [] recvMsg;
	return 1;
}

/**
 * モーターに停止指示を送信する
 * @return 成功の場合は0
*/
int CMoveWheelChair::Stop()
{
	return Move(0,0,0);
}

/**
 * 車輪モータを制御するモータドライバのサーボ制御をON、OFFする
 * @param[in] on　1でON　0でOFF
 * @return 成功の場合は0
 */
int CMoveWheelChair::Servo(int on)
{
	char com[2];
	if (on)	{ com[0] = 0x01; com[1] = 0x04; }
	else	{ com[0] = 0x00; com[1] = 0x00; }
	Send(0x400, sizeof(com), com);
	return 0;
}

/**
 * CANインターフェースにデータ送信する
 * @param[in] id   送信ID
 * @param[in] len　送信データ長
 * @param[in] data 送信データ
 * @return 成功の場合は0
 */
int CMoveWheelChair::Send(int id, int len, char *data)
{
#ifdef IFCAN
	CAN_MESSAGE		sendMessage;

	sendMessage.ulLength = len;		// Data length
	sendMessage.ulFlag = 0;			// Data frame
	sendMessage.ulID = id;			// ID
	memcpy(sendMessage.bData, data, len);
	sendMessage.ulTime = 0;			// Delay time: no
	// Transmit the CAN message.
	CanSendMessage(can_fd, &sendMessage, 1);
#endif

#ifdef PCAN
	TPCANMsg sendMessage;
	sendMessage.ID = id;
	sendMessage.LEN = len;
	sendMessage.MSGTYPE = CAN_INIT_TYPE_ST;
	memcpy(sendMessage.DATA, data, len);

	// Transmit the CAN message.
	if((errno = CAN_Write(can_fd, &sendMessage))) {
		perror("canSend_driver (Peak_Linux) : error of writing.\n");
		return 1;
	}
#endif

#ifdef CANFESTIVAL
	Message		sendMessage;
	sendMessage.len = len;		// Data length
	sendMessage.rtr = 0;			// Data frame
	sendMessage.cob_id = id;			// ID
	memcpy(sendMessage.data, data, len);

	//sendMessage.ulTime = 0;			// Delay time: no
	//CanSendMessage(portHandle, &sendMessage, 1);
	canSend_driver(can_fd, &sendMessage);
#endif

	return 0;
}

/**
 * CANインターフェースからデータを受信する
 * @param[out] msgCount 受信メッセージ数
 * @param[out] receiveMessage 受信メッセージデータ
 * @return 成功の場合は0
 */
int CMoveWheelChair::Recv(unsigned long *msgCount, CanMessage *receiveMessage)
{
#ifdef IFCAN
	CAN_PORT_STATUS	portStatus;
	CAN_MESSAGE recvMsg;

	// Wait for the data receiving.
	CanGetStatus(can_fd, &portStatus);
	if(portStatus.ulRXBCount >= 1){
		// Retrieve the CAN message.
		*msgCount = portStatus.ulRXBCount;
		CanReceiveMessage(can_fd, &recvMsg, msgCount);
		if(*msgCount > 0){
			for(unsigned int i=0; i < *msgCount; ++i){
				receiveMessage[i].len = recvMsg.ulLength;
				receiveMessage[i].cob_id = recvMsg.ulID;
				memcpy(receiveMessage[i].data, recvMsg.bData, recvMsg.ulLength);
			}

			//printf("MsgCount:%d\n",(int)*msgCount);
			return 1;
		}
		return 1;
	}
#endif

#ifdef CANFESTIVAL
	*msgCount = 1;
	Message recvMsg;

	if(1==canReceive_driver(can_fd, &recvMsg)){
		receiveMessage->cob_id = recvMsg.cob_id;
		receiveMessage->len = recvMsg.len;
		memcpy(receiveMessage->data, recvMsg.data, receiveMessage->len);

		return 0;
	}
#endif

#ifdef PCAN
	TPCANRdMsg msg;
	int nPendingReads=0, nPendingSends=0;
	// Wait for the data receiving.
	LINUX_CAN_Extended_Status(can_fd, &nPendingReads, &nPendingSends);

	if(nPendingReads > RECV_BUFF_SIZE) {
		*msgCount = RECV_BUFF_SIZE;
	}
	else {
		*msgCount = nPendingReads;
	}

	if(*msgCount > 0){
		for(unsigned int i=0; i < *msgCount; ++i){
			LINUX_CAN_Read_Timeout(can_fd, &msg, 0);
			receiveMessage[i].len = msg.Msg.LEN;
			receiveMessage[i].cob_id = msg.Msg.ID;
			for(int j=0; j < msg.Msg.LEN; ++j){
				receiveMessage[i].data[j] = msg.Msg.DATA[j];
			}
		}

		//printf("MsgCount:%d\n",(int)*msgCount);
		return 0;
	}
#endif

	return 1;

}
